import os

from .exceptions.exceptions import InvalidDaemonAddressException

DAEMON_ADDRESS_KEY = "AWS_XRAY_DAEMON_ADDRESS"
DEFAULT_ADDRESS = '127.0.0.1:2000'


class DaemonConfig(object):
    """The class that stores X-Ray daemon configuration about
    the ip address and port for UDP and TCP port. It gets the address
    string from ``AWS_TRACING_DAEMON_ADDRESS`` and then from recorder's
    configuration for ``daemon_address``.
    A notation of '127.0.0.1:2000' or 'tcp:127.0.0.1:2000 udp:127.0.0.2:2001'
    are both acceptable. The former one means UDP and TCP are running at
    the same address.
    By default it assumes a X-Ray daemon running at 127.0.0.1:2000
    listening to both UDP and TCP traffic.
    """
    def __init__(self, daemon_address=DEFAULT_ADDRESS):
        if daemon_address is None:
            daemon_address = DEFAULT_ADDRESS

        val = os.getenv(DAEMON_ADDRESS_KEY, daemon_address)
        configs = val.split(' ')
        if len(configs) == 1:
            self._parse_single_form(configs[0])
        elif len(configs) == 2:
            self._parse_double_form(configs[0], configs[1], val)
        else:
            raise InvalidDaemonAddressException('Invalid daemon address %s specified.' % val)

    def _parse_single_form(self, val):
        try:
            configs = val.split(':')
            self._udp_ip = configs[0]
            self._udp_port = int(configs[1])
            self._tcp_ip = configs[0]
            self._tcp_port = int(configs[1])
        except Exception:
            raise InvalidDaemonAddressException('Invalid daemon address %s specified.' % val)

    def _parse_double_form(self, val1, val2, origin):
        try:
            configs1 = val1.split(':')
            configs2 = val2.split(':')
            mapping = {
                configs1[0]: configs1,
                configs2[0]: configs2,
            }

            tcp_info = mapping.get('tcp')
            udp_info = mapping.get('udp')

            self._tcp_ip = tcp_info[1]
            self._tcp_port = int(tcp_info[2])
            self._udp_ip = udp_info[1]
            self._udp_port = int(udp_info[2])
        except Exception:
            raise InvalidDaemonAddressException('Invalid daemon address %s specified.' % origin)

    @property
    def udp_ip(self):
        return self._udp_ip

    @property
    def udp_port(self):
        return self._udp_port

    @property
    def tcp_ip(self):
        return self._tcp_ip

    @property
    def tcp_port(self):
        return self._tcp_port



import copy
import wrapt

from aws_xray_sdk.core import xray_recorder


class XRayTracedConn(wrapt.ObjectProxy):

    _xray_meta = None

    def __init__(self, conn, meta={}):

        super(XRayTracedConn, self).__init__(conn)
        self._xray_meta = meta

    def cursor(self, *args, **kwargs):

        cursor = self.__wrapped__.cursor(*args, **kwargs)
        return XRayTracedCursor(cursor, self._xray_meta)


class XRayTracedCursor(wrapt.ObjectProxy):

    _xray_meta = None

    def __init__(self, cursor, meta={}):

        super(XRayTracedCursor, self).__init__(cursor)
        self._xray_meta = meta

        # we preset database type if db is framework built-in
        if not self._xray_meta.get('database_type'):
            db_type = cursor.__class__.__module__.split('.')[0]
            self._xray_meta['database_type'] = db_type

    def __enter__(self):

        value = self.__wrapped__.__enter__()
        if value is not self.__wrapped__:
            return value
        return self

    @xray_recorder.capture()
    def execute(self, query, *args, **kwargs):

        add_sql_meta(self._xray_meta)
        return self.__wrapped__.execute(query, *args, **kwargs)

    @xray_recorder.capture()
    def executemany(self, query, *args, **kwargs):

        add_sql_meta(self._xray_meta)
        return self.__wrapped__.executemany(query, *args, **kwargs)

    @xray_recorder.capture()
    def callproc(self, proc, args):

        add_sql_meta(self._xray_meta)
        return self.__wrapped__.callproc(proc, args)


def add_sql_meta(meta):

    subsegment = xray_recorder.current_subsegment()

    if not subsegment:
        return

    if meta.get('name', None):
        subsegment.name = meta['name']

    sql_meta = copy.copy(meta)
    if sql_meta.get('name', None):
        del sql_meta['name']
    subsegment.set_sql(sql_meta)
    subsegment.namespace = 'remote'



import wrapt
import sqlite3

from aws_xray_sdk.ext.dbapi2 import XRayTracedConn


def patch():

    wrapt.wrap_function_wrapper(
        'sqlite3',
        'connect',
        _xray_traced_connect
    )


def _xray_traced_connect(wrapped, instance, args, kwargs):

    conn = wrapped(*args, **kwargs)

    meta = {}
    meta['name'] = args[0]
    meta['database_version'] = sqlite3.sqlite_version

    traced_conn = XRayTracedSQLite(conn, meta)

    return traced_conn


class XRayTracedSQLite(XRayTracedConn):

    def execute(self, *args, **kwargs):
        return self.cursor().execute(*args, **kwargs)

    def executemany(self, *args, **kwargs):
        return self.cursor().executemany(*args, **kwargs)



import threading
from enum import Enum


class Reservoir(object):
    """
    Centralized thread-safe reservoir which holds fixed sampling
    quota, borrowed count and TTL.
    """
    def __init__(self):
        self._lock = threading.Lock()

        self._quota = None
        self._TTL = None

        self._this_sec = 0
        self._taken_this_sec = 0
        self._borrowed_this_sec = 0

        self._report_interval = 1
        self._report_elapsed = 0

    def borrow_or_take(self, now, can_borrow):
        """
        Decide whether to borrow or take one quota from
        the reservoir. Return ``False`` if it can neither
        borrow nor take. This method is thread-safe.
        """
        with self._lock:
            return self._borrow_or_take(now, can_borrow)

    def load_quota(self, quota, TTL, interval):
        """
        Load new quota with a TTL. If the input is None,
        the reservoir will continue using old quota until it
        expires or has a non-None quota/TTL in a future load.
        """
        if quota is not None:
            self._quota = quota
        if TTL is not None:
            self._TTL = TTL
        if interval is not None:
            self._report_interval = interval / 10

    @property
    def quota(self):
        return self._quota

    @property
    def TTL(self):
        return self._TTL

    def _time_to_report(self):
        if self._report_elapsed + 1 >= self._report_interval:
            self._report_elapsed = 0
            return True
        else:
            self._report_elapsed += 1

    def _borrow_or_take(self, now, can_borrow):
        self._adjust_this_sec(now)
        # Don't borrow if the quota is available and fresh.
        if (self._quota is not None and self._quota >= 0 and
                self._TTL is not None and self._TTL >= now):
            if(self._taken_this_sec >= self._quota):
                return ReservoirDecision.NO

            self._taken_this_sec = self._taken_this_sec + 1
            return ReservoirDecision.TAKE

        # Otherwise try to borrow if the quota is not present or expired.
        if can_borrow:
            if self._borrowed_this_sec >= 1:
                return ReservoirDecision.NO

            self._borrowed_this_sec = self._borrowed_this_sec + 1
            return ReservoirDecision.BORROW

    def _adjust_this_sec(self, now):
        if now != self._this_sec:
            self._taken_this_sec = 0
            self._borrowed_this_sec = 0
            self._this_sec = now


class ReservoirDecision(Enum):
    """
    An Enum of decisions the reservoir could make based on
    assigned quota with TTL and the current timestamp/usage.
    """
    TAKE = 'take'
    BORROW = 'borrow'
    NO = 'no'



import threading

from .reservoir import Reservoir
from aws_xray_sdk.core.utils.search_pattern import wildcard_match


class SamplingRule(object):
    """
    Data model for a single centralized sampling rule definition.
    """
    def __init__(self, name, priority, rate, reservoir_size,
                 host=None, method=None, path=None, service=None,
                 service_type=None):
        self._name = name
        self._priority = priority
        self._rate = rate
        self._can_borrow = not not reservoir_size

        self._host = host
        self._method = method
        self._path = path
        self._service = service
        self._service_type = service_type

        self._reservoir = Reservoir()
        self._reset_statistics()

        self._lock = threading.Lock()

    def match(self, sampling_req):
        """
        Determines whether or not this sampling rule applies to the incoming
        request based on some of the request's parameters.
        Any ``None`` parameter provided will be considered an implicit match.
        """
        if sampling_req is None:
            return False

        host = sampling_req.get('host', None)
        method = sampling_req.get('method', None)
        path = sampling_req.get('path', None)
        service = sampling_req.get('service', None)
        service_type = sampling_req.get('service_type', None)

        return (not host or wildcard_match(self._host, host)) \
            and (not method or wildcard_match(self._method, method)) \
            and (not path or wildcard_match(self._path, path)) \
            and (not service or wildcard_match(self._service, service)) \
            and (not service_type or wildcard_match(self._service_type, service_type))

    def is_default(self):
        # ``Default`` is a reserved keyword on X-Ray back-end.
        return self.name == 'Default'

    def snapshot_statistics(self):
        """
        Take a snapshot of request/borrow/sampled count for reporting
        back to X-Ray back-end by ``TargetPoller`` and reset those counters.
        """
        with self._lock:

            stats = {
                'request_count': self.request_count,
                'borrow_count': self.borrow_count,
                'sampled_count': self.sampled_count,
            }

            self._reset_statistics()
            return stats

    def merge(self, rule):
        """
        Migrate all stateful attributes from the old rule
        """
        with self._lock:
            self._request_count = rule.request_count
            self._borrow_count = rule.borrow_count
            self._sampled_count = rule.sampled_count
            self._reservoir = rule.reservoir
            rule.reservoir = None

    def ever_matched(self):
        """
        Returns ``True`` if this sample rule has ever been matched
        with an incoming request within the reporting interval.
        """
        return self._request_count > 0

    def time_to_report(self):
        """
        Returns ``True`` if it is time to report sampling statistics
        of this rule to refresh quota information for its reservoir.
        """
        return self.reservoir._time_to_report()

    def increment_request_count(self):
        with self._lock:
            self._request_count += 1

    def increment_borrow_count(self):
        with self._lock:
            self._borrow_count += 1

    def increment_sampled_count(self):
        with self._lock:
            self._sampled_count += 1

    def _reset_statistics(self):
        self._request_count = 0
        self._borrow_count = 0
        self._sampled_count = 0

    @property
    def rate(self):
        return self._rate

    @rate.setter
    def rate(self, v):
        self._rate = v

    @property
    def name(self):
        return self._name

    @property
    def priority(self):
        return self._priority

    @property
    def reservoir(self):
        return self._reservoir

    @reservoir.setter
    def reservoir(self, v):
        self._reservoir = v

    @property
    def can_borrow(self):
        return self._can_borrow

    @property
    def request_count(self):
        return self._request_count

    @property
    def borrow_count(self):
        return self._borrow_count

    @property
    def sampled_count(self):
        return self._sampled_count



from .recorder import AWSXRayRecorder
from .patcher import patch_all, patch
from .utils.compat import PY35


if not PY35:
    xray_recorder = AWSXRayRecorder()
else:
    from .async_recorder import AsyncAWSXRayRecorder
    xray_recorder = AsyncAWSXRayRecorder()

__all__ = [
    'patch',
    'patch_all',
    'xray_recorder',
    'AWSXRayRecorder',
]
